//
// VFO for IC Based Receiver
//
// Based on a port to PIC of Hans Summers' Si5351a Arduino demo 
// Ported by Andy Fletcher, 2016-07-28 in Edinburgh, Scotland.
// Changes for PIC12F1840 2021-06-16 by Andrew Fletcher
//
// The Si5351A part pulled out into a header for easier insertion into other projects
//
// Original Author: Hans Summers, 2015
// Website: http://www.hanssummers.com
//
// Uses the Si5351a module kit http://www.hanssummers.com/synth
// Please also refer to SiLabs AN619 which describes all the registers to use
//

// TODO
//		None

//
// Application Configuration
//
// Minimum frequency in Hz
#define MIN_FREQ 7000000
// Maximum frequency in Hz
#define MAX_FREQ 7020000
// Frequency change per encoder click in Hz, max 65536
// Note: some encoders seem to 'click' twice per detente
#define FREQ_SLOW_DELTA 10
#define FREQ_FAST_DELTA 500
// !!!! REMEMBER TO TRIM THE SI_XTAL_FREQ IN SI5351A.H !!!!
// set SWAP_CW to 1 to reverse clockwise and anti-clockwise as it depends on the wiring
// set SWAP_CW to 0 to do no swapping
#define SWAP_CW 1
// Length of poll delay to throttle number of frequency changes. Must be > 0. 
#define POLL_DELAY 1
// end of application configuration

// #define USE_LED_FOR_DEBUG 1

// Configuration bits (PIC12F1840) 
#pragma config FCMEN = ON						// enable fail-safe clock monitor
#pragma config IESO = ON 						// ??
#pragma config CLKOUTEN = OFF                   // disable oscillator
#pragma config BOREN = ON						// enable brown-out detection	
#pragma config CPD = OFF						// disable data EEPROM code protection
#pragma config CP = OFF							// disable code protection
#pragma config MCLRE = OFF						// MCLR is digital input with weak pullup
#pragma config PWRTE = OFF						// disable power-up timer
#pragma config WDTE = OFF						// disable watchdog
#pragma config FOSC = INTOSC					// but disabled by CLKOUTEN
#pragma config LVP = OFF						// works with MCLRE

#include <xc.h>
#include <stdint.h>
#include "i2c.h"
#define I2C_SLAVE 0x60                    // Slave device I2C address needed by si5351a.h
#include "si5351a.h"

//
// LED definitions
//
#define FREQ_LIMIT_LED PORTAbits.RA0

//
// Encoder definitions
// 
// encoder rotation use RA4 and RA5
#define ENCODER         ((PORTA & 0x30) >> 4)
#define ENCODER_BUTTON  PORTAbits.RA3


//
// Global variables
//
volatile unsigned char encoderTurned;
volatile unsigned char encoderBits, oldEncoderBits;
volatile unsigned char cw; 		// clockwise indicator for encoder
volatile uint32_t delta;
volatile uint32_t frequency;

//
// ISR
//
void interrupt handleInterrupt(void) {

  //
  // handle encoder interrupts
  // 
  if (INTCONbits.IOCIF) {
// #ifdef USE_LED_FOR_DEBUG
// FREQ_LIMIT_LED = 1;
// #endif
		//
		// get latest state of encoder bits
		//
		encoderBits = ENCODER & 0b11;

		// 
		// handle in state machine, ignore 'impossible' changes
		//
		
		// When travelling clockwise the encoder bits go     
		// 11 -> 10 -> 00 -> 01 -> 11
		//
		// When travelling anticlockwise the encoder bits go 
		// 11 -> 01 -> 00 -> 10 -> 11
		//
		switch (oldEncoderBits) {
          case 0b00: 
            if (encoderBits == 0b01) cw = 1;  // encoder is turning clockwise
            else
            if (encoderBits == 0b10) cw = 0;  // encoder is turning anticlockwise
            break;
          case 0b01:
            if ((encoderBits == 0b11) && (cw == 1)) { 
              // encoder has turned a notch clockwise
              encoderTurned = 1;
            }
            else
            if (encoderBits == 0b00) cw = 0;  // encoder is turning anticlockwise
            break;
          case 0b10:
            if (encoderBits == 0b00) cw = 1;  // encoder is turning clockwise
            else
            if ((encoderBits == 0b11) && (cw == 0)) { 
              // encoder has turned a notch anticlockwise
              encoderTurned = 1;
            }
            break;
          case 0b11:
            if (encoderBits == 0b10) cw = 1;  // encoder is turning clockwise
            else
            if (encoderBits == 0b01) cw = 0;  // encoder is turning anticlockwise
            break;
		}
		
		oldEncoderBits = encoderBits; 	// store for next interrupt

		if (encoderTurned != 0) {				// the encoder was turned
			delta = (ENCODER_BUTTON ? FREQ_SLOW_DELTA : FREQ_FAST_DELTA); // fast tune if button is down
#ifdef SWAP_CW
				cw = (cw == 0) ? 1:0;
#endif 
			if (cw == 1) { 						
				frequency += delta;
				frequency = (frequency > MAX_FREQ ? MAX_FREQ : frequency); // apply limit
			} else {
				frequency -= delta;
				frequency = (frequency < MIN_FREQ ? MIN_FREQ : frequency); // apply limit
			}

#ifndef USE_LED_FOR_DEBUG
			if ((frequency == MIN_FREQ) || (frequency == MAX_FREQ)) {
				FREQ_LIMIT_LED = 1; 			// indicate limit reached
			} else {
				FREQ_LIMIT_LED = 0;
			}
#endif
		 }
// #ifdef USE_LED_FOR_DEBUG
// FREQ_LIMIT_LED = 0; 
// #endif
		// clear whichever interrupt-on-change flag was set 
    	if (IOCAF4) {
    		IOCAF &= 0x2f; 				// clear the IOCAF4 flag
    	} else if (IOCAF5) {
    		IOCAF &= 0x1f;				// clear the IOCAF5 flag
    	}
  }
}

//
// production functions
//

void setClock(void) {
  // set PIC internal oscillator to 32MHz
  OSCCONbits.IRCF0 = OSCCONbits.IRCF1 = OSCCONbits.IRCF2 = 1;
  OSCCONbits.SPLLEN = 1;
}

// execution
void main(void)
{
  unsigned int pollDelay;

  // disable all interrupts
  INTCON = 0;

  // configure i/o ports
  ANSELA = 0;            	// configure all as digital unless needed otherwise

  CM1CON0 = CM1CON1 = 0;   	// disable comparators

  TRISA = 0xFF;		    	// all of PORTA are inputs
  TRISAbits.TRISA0 = 0;	    // except frequency limit LED

  OPTION_REG = 0b01011110;	// enable pull-ups, PS = 1:128, NOT TMR0
  WPUA = 0b111110;			// weak pull-ups on I2C and encoder pins

  setClock();

  //
  // initialise the encoder
  //
  oldEncoderBits = ENCODER;
  IOCAP = 0;				// interrupt on encoder turn change
  IOCAPbits.IOCAP4 = IOCAPbits.IOCAP5 = 1;		// get interrupted if RA4 or RA5 go high
  IOCANbits.IOCAN4 = IOCANbits.IOCAN5 = 1;		// falling edge too
  encoderTurned = 0;							// set in ISR

  //
  // initialise the clock generator
  //
  i2c_Init();
  si5351aOutputOff(SI_CLK1_CONTROL);
  si5351aOutputOff(SI_CLK2_CONTROL);

  //
  // initialise LED
  //
  FREQ_LIMIT_LED = 0;

  // set the initial frequency to halfway between the frequency limits
  frequency = MIN_FREQ + ((MAX_FREQ - MIN_FREQ) / 2);
  si5351aSetFrequency(frequency);

  delta = FREQ_SLOW_DELTA; // start with slow tuning

  INTCONbits.IOCIE = INTCONbits.GIE = 1;		// allow encoder interrupts 
  
  /* loop forever */
  while (1) {
 	 INTCONbits.IOCIE = INTCONbits.GIE = 0;	// handle turns with interrupts disabled
	 if (encoderTurned != 0) {				// the encoder was turned
#ifdef USE_LED_FOR_DEBUG
FREQ_LIMIT_LED = 1; 
#endif
		si5351aSetFrequency(frequency);		// can't be done in ISR as it is non-reentrant
		encoderTurned = 0;					// reset turn flag
	 }
	 INTCONbits.IOCIE = INTCONbits.GIE = 1;	// re-enable encoder interrupts
	 
	 pollDelay = POLL_DELAY; while(pollDelay--); // allow time for turns to be seen
	 
	 // sleep until interrupt
	 #asm
	 SLEEP
	 #endasm								
  }
}

